# studio5.r
# NEEDS TO BE ORGANIZED TO RUN!

# K-Sided Dice Example from Lecture 11

# Random process: choose among K-sided Dies at random and roll the die
# Conduct Bayesian analysis inferring the chosen die type using outcomes of sample rolls

#--------------------------------
# Example 1: UPDATING FOR ONE ROLL
#--------------------------------
# 4,6,8,12,20 sided dice, all equally likely to be chosen
dice=c(4,6,8,12,20)
prior=c(.2,.2,.2,.2,.2)

# Build likelihood table 5 dice x 20 possible outcomes
# First create the table with all 0's
likelihoodTable = matrix(0,nrow=5, ncol=20)
# Make the probabilities (rows) for each die separately.
# (We could have done this in a loop)
# 4-sided die; outcomes 1-4 are set to 1/4;
# outcomes 5-20 are left at 0. Other dice are similar
likelihoodTable[1,1:4] = 1/4
likelihoodTable[2,1:6] = 1/6
likelihoodTable[3,1:8] = 1/8
likelihoodTable[4,1:12] = 1/12
likelihoodTable[5,1:20] = 1/20
print(likelihoodTable, digits=3)

title = "Prior probabilities"
barplot(prior, col='blue', width=rep(.1,5),
        xlim=c(0,5),space=3,
        names=dice, main=title)
nrolls=20
studioBayesFunction(dice,prior,likelihoodTable,nrolls)
# Choose die according to prior distribution
set.seed(3)  # set.seed allows us to get reproducible 'random' data
randomDie=sample(dice,1,prob=prior)

# roll the die once
x = sample(1:randomDie, size=1, replace=TRUE)
print(x)

# Choose Likelihood column:
likelihood = likelihoodTable[,x]
print(likelihood)

bayesNumerator = prior*likelihood
posterior = bayesNumerator/sum(bayesNumerator)

title = paste("Posterior probabilities after one roll: outcome =", x)
barplot(posterior, col='orange', width=rep(.1,5),
        xlim=c(0,5),space=3,
        names=dice, main=title)

#-------------------------------------
# Storing the entire table in a data frame

# Put the Bayes table in a data frame for easy printing.
# Data frames have named columns, the syntax is name = values.
# So dice = dice creates a column named 'dice' with values from the dice vector.
# (They don't need to be the same name, e.g.
# 'platonicSolids' = dice would create a column
# called 'platonicSolids' with values from the dice
# vector.)
#
bayesTable=data.frame(
    dice=dice,
    prior=prior,
    likelihood=likelihood,
    bayesNumer=bayesNumerator,
    posterior=posterior
)
title = paste("Bayes table after one roll: roll =", x)
print(title)
print(bayesTable)

#--------------------------------
# Example 2: ROLL THE CHOSEN DIE REPEATEDLY
#--------------------------------
# Just to keep this part self-contained we repeat the
# setup above
nrolls=20  #roll the die nrolls times
dice=c(4,6,8,12,20)
prior=c(.2,.2,.2,.2,.2)
# Build likelihood table 5 dice x 20 possible outcomes
# First create the table with all 0s
likelihoodTable = matrix(0,nrow=5, ncol=20)
# Make the probabilities (rows) for each die separately.
# (We could have done this in a loop)
# 4-sided die; outcomes 1-4 are set to 1/4;
# outcomes 5-20 are left at 0. Other dice are similar
likelihoodTable[1,1:4] = 1/4
likelihoodTable[2,1:6] = 1/6
likelihoodTable[3,1:8] = 1/8
likelihoodTable[4,1:12] = 1/12
likelihoodTable[5,1:20] = 1/20

# Choose die according to prior distribution
set.seed(1)
randomDie=sample(dice,1,prob=prior)

# Get all the data at once
dataRolls = sample(1:randomDie, size=nrolls,
                   replace=TRUE)

# It's good practice to just plot the data to see if there are obvious problems
plot(dataRolls, xlab="Roll Index", main="Sample of iid Rolls")

# Plot the prior
title = "Prior probabilities"
barplot(prior, col='blue', width=rep(.1,5),xlim=c(0,5),
        space=3, names=dice, main=title)

# Create the data frame to hold  our computations
bayesTable=data.frame(
    dice=dice,
    prior=NA,
    likelihood=NA,
    bayesNumerator=NA,
    posterior=NA)

# Initialize matrix whose jth column will store
# the posterior distribution after updating the jth roll
# We will use this to make a nifty stacked bar plot at the end
posteriorMat = matrix(NA, nrow=length(dice), ncol=nrolls )

#set the first prior
prior.jroll = prior

# Go throught the updata process for each roll
for (jroll in 1:nrolls){
  x.jroll =dataRolls[jroll]

  likelihood.jroll = likelihoodTable[,x.jroll]
  bayesNumerator.jroll = prior.jroll * likelihood.jroll
  posterior.jroll = bayesNumerator.jroll/sum(bayesNumerator.jroll)

  # store the posterior
  posteriorMat[,jroll] = posterior.jroll

  # plot the posterior
  title = paste("Posterior prob. after roll ", jroll, ": outcome =", x.jroll)
  barplot(posterior.jroll, col='orange', width=rep(.1,5),
          xlim=c(0,5),space=3, names=dice, main=title)

  # put the table in our data frame for easy printing
  bayesTable[,"prior"] = prior.jroll
  bayesTable[,"likelihood"] = likelihood.jroll
  bayesTable[,"bayesNumerator"] = bayesNumerator.jroll
  bayesTable["posterior"] = posterior.jroll

  title = paste("Bayes table after roll", jroll, ": outcome =", x.jroll)
  print(title)
  print(bayesTable)

  # SET THE PRIOR for the next roll
  prior.jroll = posterior.jroll
}
# NOTE: There should be a sequence of barplots -one for each roll-- in the plot window

# stacked barplot of the prior/posterior distributions
# as a function of the number of rolls
# (cbind --column bind is a easy way to add columns to a matrix)
allProbs = cbind(prior,posteriorMat)
barplot(allProbs,legend.text=paste("D",dice,sep=""),
        names.arg=c("0(Prior)",c(1:nrolls)),
        col=rainbow(length(dice)))
title(xlab="Number of Rolls")
title(main="Stacked Barplot of Posterior (Prior) Probabilities")
# sub-bar heights equal posterior probabilities
####################################

#----------------------------
# Avoiding repeated code by writing a function. Here is a function
# that does what the code above does; you can paste it once at the
# top of your solution instead of recopying long things many times.
# Each different kind of updating question will require you to
# make new prior, likelihood table, and datarolls; after doing
# that, a single line like

# studioBayesFunction(dice, prior, standardLikelihoodTable, dataRolls)

# will do most of the work.

# This code does Bayesian updating for the probability a 
# certain die was chosen from the list of dice. The arguments are:
#  dice = list of dice (hypotheses)
#  prior = prior probabilities on the hypotheses
#  likelihoodTable -- each row is the probability table
#  for a hypothesis
#  dataRolls -- random data for rolling a randomly chosen die
#  showIntermediateResults is a boolean to print tables and draw 
# barplots of intermediate posteriors, it defaults to TRUE. 
# Setting it to FALSE can significantly speed up the code.
#  isCensoredData is a boolean used for problem 3. It says the 
# data has been 'censored'

# RED_FLAG: we don't check compatibility between the arguments
# 1. length(dice) == length(prior) == number of rows of likelihoodTable
# 2. Allowable data = 1 to number of columns of  likelihoodTable

studioBayesFunction = function(dice,
                               prior,
                               likelihoodTable,
                               dataRolls,
                               showIntermediateResults = TRUE,
                               isCensoredData = FALSE) {
  
  nrolls = length(dataRolls)
  
  if (showIntermediateResults)
  {
    # It's good practice to just plot the data
    plot(dataRolls, xlab="Roll Index", main="Sample of iid Rolls")
    
    # Plot the prior
    title = "Prior probabilities"
    barplot(prior, col='blue', width=rep(.1,5),xlim=c(0,5),space=3,
            names=dice, main=title)
  }
  
  # Create the data frame to hold  our computations
  bayesTable=data.frame(
    dice=dice,
    prior=NA,
    likelihood=NA,
    bayesNumerator=NA,
    posterior=NA)
  
  # Initialize matrix whose jth column will store the posterior distribution after updating the jth roll. We will use this to make a nifty stacked bar plot at the end
  posteriorMat<-matrix(NA, nrow=length(dice), ncol=nrolls )
  
  #set the first prior
  prior.jroll = prior
  
  # Go throught the updata process for each roll
  for (jroll in 1:nrolls) {
    x.jroll =dataRolls[jroll]
    
    if (isCensoredData)
    {
      # censored data = 0 corresponds to column 1 in the likelihood table censored data = 1 corresponds to column 2 in the likelihood table
      likelihoodColumn = x.jroll + 1
    }
    else {
      #normally the likelihood column number is the roll
      likelihoodColumn = x.jroll
    }
    likelihood.jroll = likelihoodTable[,likelihoodColumn]
    bayesNumerator.jroll = prior.jroll * likelihood.jroll
    posterior.jroll = bayesNumerator.jroll/sum(bayesNumerator.jroll)
    
    # store the posterior
    posteriorMat[,jroll] = posterior.jroll
    
    if (showIntermediateResults)
    {
      # plot the posterior
      title = paste("Posterior probabilities after roll ", jroll, ": roll =", x.jroll)
      barplot(posterior.jroll, col='orange',
              width=rep(.1,5),
              xlim=c(0,5),space=3,
              names=dice, main=title)
      
      # put the table in our data frame for easy printing
      bayesTable[,"prior"] = prior.jroll
      bayesTable[,"likelihood"] = likelihood.jroll
      bayesTable[,"bayesNumerator"] = bayesNumerator.jroll
      bayesTable["posterior"] = posterior.jroll
      
      title = paste("Bayes table after one roll",
                    jroll, ": roll =", x.jroll)
      print(title)
      print(bayesTable)
    }
    # SET THE PRIOR for the next roll
    prior.jroll = posterior.jroll
    
  } # this is the end of the loop through the nrolls 
  # different rolls. Now plot the conclusions.
  
  # stacked barplot of the prior/posterior distributions
  # as a function of the number of rolls
  # (cbind --column bind is a easy way to add columns to a matrix)
  allProbs = cbind(prior,posteriorMat)
  barplot(allProbs,legend.text=paste("Sides",dice,sep=""),
          names.arg=c("0(Prior)",c(1:nrolls)),
          col=rainbow(length(dice)), border=NA)
  title(xlab="Number of Rolls")
  title(main="Stacked Barplot of Posterior (Prior) Probabilities")
  # sub-bar heights equal posterior probabilities
}
